# %%

import os
import platform

from . import OpenBankingBR

from .errors import BatchException
from .models import Participante, Produto

from typing import List
from datetime import date, datetime

class BatchOpenBanking():
    """
    Classe destinada para o processamento batch de ingestão dos dados públicos dos
    participantes listados no diretório do Open Banking Brasil.

    ```python 
    # exemplo para baixar os dados    
    bob = BatchOpenBanking(
        dataDir  = './data',
        cacheDir = './cache',

        # caso queira trabalhar no excel mais facilmente, só remover o comentário.
        # encoding     = 'cp1252',
        # csvDelimiter = ';'
    )
    bob.todos_dados()
    bob.limpa_cache_antigo() # remove arquivos de cache com mais de 5 dias de vida.
    ```

    """
    def __init__(self, dataDir: str, cacheDir: str, csvDelimiter: str = ',', encoding = 'utf8', ignoraErros: bool = False):
        """
        Inicializa o objeto com os parâmetros de execução do batch.

        Caso `dataDir` ou `cacheDir` não existam eles serão criados.

        Argumentos:

            dataDir      - Diretório onde será salvo os arquivos processados.
            cacheDir     - Diretório onde é salvo os arquivos de cache usados para não abusar as APIs 
                           disponibilizado pelas instituições financeiras. O comportamento do cache é
                           de fazer o download de cada endPoint uma única vez por dia.
            csvDelimiter - Especifica o delimitador do arquivo csv (caso queira usar outro fora do padrão).
            ignoreErrors - Em vez do exception uma mensagem de erro é logada e vida que segue.
        """
        if not os.path.exists(dataDir):
            os.makedirs(dataDir)
        elif not os.path.isdir(dataDir):
            raise ValueError("O valor especificado no argumento dataDir não é um diretório.")

        if not os.path.exists(cacheDir):
            os.makedirs(cacheDir)
        elif not os.path.isdir(cacheDir):
            raise ValueError("O valor especificado no argumento cacheDir não é um diretório.")

        self.dataDir        = dataDir
        self.cacheDir       = cacheDir
        self.csvDelimiter   = csvDelimiter
        self.ignoraErros    = ignoraErros
        self.encoding       = encoding

    def todos_dados(self):
        """
        Realiza o download de todos os dados públicos disponíveis dos participantes
        registrados no diretório do Open Banking Brasil, salvando os dados em arquivos
        na pasta `dataDir` informada na inicialização do `BatchOpenBanking`.

        Os arquivos serão salvos com o nome no seguinte formato: `yyyymmdd_openbanking_xxx.csv`
        """
        self.agencias_csv()
        self.produtos_csv()
        self.servicos_csv()
        self.pacotes_csv()

    def limpa_cache_antigo(self, idadeMaxima = 5):
        """
        Realiza a limpeza dos arquivos antigos (padrão é `5` ou mais dias de vida) de cache.

        Note que um arquivo de cache só deve ser antigo caso um endpoint não esteja mais
        disponível por uma financeira, e mesmo que o arquivo seja "antigo" ao tentar puxar
        os dados mais recentes, o arquivo de cache será recriado com as informações mais recentes.
        """
    
        if not os.path.exists(self.cacheDir):
            return
        
        if not os.path.isdir(self.cacheDir):
            return

        ctime: datetime
        files = os.listdir(self.cacheDir)
        for file in files:
            if not file.endswith('.json'):
                continue
           
            fileName = os.path.join(self.cacheDir, file)

            if platform.system() == "Windows":
                ctime = datetime.fromtimestamp(os.path.getctime(fileName))
            else:
                stat = os.stat(fileName)
                try:
                    ctime = datetime.fromtimestamp(stat.st_birthtime)
                except:
                    ctime = datetime.fromtimestamp(stat.st_mtime)

            if (date.today() - ctime.date()).days > idadeMaxima:
                try:
                    os.remove(fileName)
                except:
                    continue

    def agencias_csv(self):
        """
        Realiza o download de todas as agências informadas nas apis dos participantes
        no diretório do Open Banking Brasil. Os dados das agências são salvos no arquivo 
        com nome: `yyyymmdd_openbanking_agencias.csv`.
        """
        ob = OpenBankingBR(
            cacheDir = self.cacheDir,
        )

        participantes: List[Participante]
        participantes = None

        hoje = date.today().strftime('%Y%m%d')

        fileName = os.path.join(self.dataDir, f'{hoje}_openbanking_agencias.csv')
        if os.path.exists(fileName):
            os.remove(fileName)
            print(f'Arquivo {fileName} foi apagado para o reprocessamento.')

        with open(fileName, 'w', encoding=self.encoding) as file:

            file.write(self.csvDelimiter.join([
                "DATA_BASE",
                "API",
                "PARTICIPANTE_SEQ",
                "PARTICIPANTE_CNPJ",
                "PARTICIPANTE_NOME",
                "AGENCIA_SEQ",
                "AGENCIA_TIPO",
                "AGENCIA_CODIGO",
                "AGENCIA_DIGITO",
                "AGENCIA_NOME",
                "AGENCIA_TELEFONE",
                "AGENCIA_ENDERECO",
                "AGENCIA_COMPLEMENTO",
                "AGENCIA_BAIRRO",
                "AGENCIA_CIDADE",
                "AGENCIA_UF",
                "AGENCIA_CEP",
                "AGENCIA_CODIGO_IBGE",
                "AGENCIA_LATITUDE",
                "AGENCIA_LONGITUDE",
            ]) + '\n')

            try:
                participantes = [x for x in ob.participantes]
            except:
                if self.ignoraErros:
                    print('Erro: Não foi possível baixar as informações do diretório de participantes do Open Banking Brasil.')
                    return

                raise BatchException("Não foi possível baixar as informações do diretório de participantes do Open Banking Brasil.")

            totalParticipantes = len(participantes)
            totalAgencias = 0

            seqParticipante = 0
            seqAgencia = 0

            for participante in participantes:
                seqAgencia = 0

                seqParticipante += 1
                try:
                    for agencia in participante.agencias:
                        totalAgencias += 1
                        seqAgencia += 1
                        row = [
                            str(hoje),
                            _fix(agencia.endPoint),
                            _fix(seqParticipante),
                            _fix(participante.cnpj),
                            _fix(participante.nome),
                            _fix(seqAgencia),
                            _fix(agencia.tipo),
                            _fix(agencia.codigo),
                            _fix(agencia.digitoVerificador),
                            _fix(agencia.nome),
                            _fix(agencia.telefone),
                            _fix(agencia.endereco),
                            _fix(agencia.enderecoComplemento),
                            _fix(agencia.bairro),
                            _fix(agencia.cidade),
                            _fix(agencia.uf),
                            _fix(agencia.cep),
                            _fix(agencia.codigoIBGE),
                            _fix(agencia.latitude),
                            _fix(agencia.longitude),
                        ]
                        file.write(self.csvDelimiter.join(row) + '\n')

                        continue

                except Exception as e:
                    if self.ignoraErros:
                        print(f"Erro: Participanete ignorado devido a erro no processamento: {e}")
                        continue 
                    
                    raise BatchException(f'Erro: Participanete ignorado devido a erro no processamento: {e}')

            file.flush()

        print('Processamento das agências concluído.')
        print(f"Total de participantes: {totalParticipantes}")
        print(f"Total de agências: {totalAgencias}")
        print('.')

    def produtos_csv(self):
        """
        Realiza o download de todos os produtos disponíveis em todos participantes 
        no diretório do Open Banking Brasil, e salva os dados dos produtos em um 
        arquivo no formato: `yyyymmdd_openbanking_produtos.csv`.
        """

        ob = OpenBankingBR(
            cacheDir = self.cacheDir,
        )

        participantes: List[Participante]
        participantes = None

        hoje = date.today().strftime('%Y%m%d')

        fileName = os.path.join(self.dataDir, f'{hoje}_openbanking_produtos.csv')
        if os.path.exists(fileName):
            os.remove(fileName)
            print(f'Arquivo {fileName} foi apagado para o reprocessamento.')
        
        with open(fileName, 'w', encoding=self.encoding) as file:
            file.write(self.csvDelimiter.join([
                "DATA_BASE",
                "API",
                "PARTICIPANTE_SEQ",
                "PARTICIPANTE_CNPJ",
                "PARTICIPANTE_NOME",
                "PRODUTO_SEQ",
                "PRODUTO_TIPO",
                "PRODUTO_CATEGORIA",
                "PRODUTO_REDE",
                "PRODUTO_NOME",
                "PRODUTO_INDEXADOR",
                "PRODUTO_INDEXADOR_RATE",
                "PRODUTO_TAXA_MINIMA",
                "PRODUTO_TAXA_MAXIMA",
                "PRODUTO_FAIXA1_TAXA",
                "PRODUTO_FAIXA1_CLIENTES",
                "PRODUTO_FAIXA2_TAXA",
                "PRODUTO_FAIXA2_CLIENTES",
                "PRODUTO_FAIXA3_TAXA",
                "PRODUTO_FAIXA3_CLIENTES",
                "PRODUTO_FAIXA4_TAXA",
                "PRODUTO_FAIXA4_CLIENTES",
                "PRODUTO_PROGRAMA_RECOMPENSAS",
            ]) + '\n')

            try:
                participantes = [x for x in ob.participantes]
            except:
                msg = "Não foi possível baixar as informações do diretório de participantes do Open Banking Brasil."
                if self.ignoraErros:
                    print(f'Erro: {msg}')
                    return

                raise BatchException(msg)

            totalParticipantes = len(participantes)
            totalProdutos = 0

            seqParticipante = 0
            seqProduto = 0

            for participante in participantes:
                seqProduto = 0

                seqParticipante += 1
                try:
                    for produto in participante.produtos:
                        seqProduto += 1
                        totalProdutos += 1

                        row = [
                            str(hoje),
                            _fix(produto.endPoint),
                            _fix(seqParticipante),
                            _fix(participante.cnpj),
                            _fix(participante.nome),
                            _fix(seqProduto),
                            _fix(produto.tipo),
                            _fix(produto.categoria),
                            _fix(produto.rede),
                            _fix(produto.nome),
                            _fix(produto.indexador),
                            _fix(produto.indexadorRate),
                            _fix(produto.taxaMinima),
                            _fix(produto.taxaMaxima),
                            _fix(produto.faixa1Taxa),
                            _fix(produto.faixa1Clientes),
                            _fix(produto.faixa2Taxa),
                            _fix(produto.faixa2Clientes),
                            _fix(produto.faixa3Taxa),
                            _fix(produto.faixa3Clientes),
                            _fix(produto.faixa4Taxa),
                            _fix(produto.faixa4Clientes),
                            _fix(produto.programaRecompensas),
                        ]
                        file.write(self.csvDelimiter.join(row) + '\n')
                        continue

                except:
                    if self.ignoraErros:
                        print('Participanete ignorado devido a erro no processamento.')
                        continue

                    raise BatchException("Um erro ocorreu durante o processamento de um participante.")

            file.flush()

        print('Processamento dos produtos concluído.')
        print(f"Total de participantes: {totalParticipantes}")
        print(f"Total de produtos: {totalProdutos}")
        print('.')

    def servicos_csv(self):
        """
        Realiza o download de todos os produtos disponíveis em todos participantes 
        no diretório do Open Banking Brasil, e salva os dados dos serviços em um 
        arquivo no formato: `yyyymmdd_openbanking_servicos.csv`.
        """

        ob = OpenBankingBR(
            cacheDir = self.cacheDir,
        )

        participantes: List[Participante]
        participantes = None

        hoje = date.today().strftime('%Y%m%d')

        fileName = os.path.join(self.dataDir, f'{hoje}_openbanking_servicos.csv')
        if os.path.exists(fileName):
            os.remove(fileName)
            print(f'Arquivo {fileName} foi apagado para o reprocessamento.')
        
        with open(fileName, 'w', encoding=self.encoding) as file:
            file.write(self.csvDelimiter.join([
                "DATA_BASE",
                "API",
                "PARTICIPANTE_SEQ",
                "PARTICIPANTE_CNPJ",
                "PARTICIPANTE_NOME",
                "PRODUTO_SEQ",
                "PRODUTO_TIPO",
                "PRODUTO_CATEGORIA",
                "PRODUTO_NOME",
                "SERVICO_SEQ",
                "SERVICO_NOME",
                "SERVICO_CODIGO",
                "SERVICO_TAXA_MINIMA",
                "SERVICO_TAXA_MAXIMA",
                "SERVICO_FAIXA1_TAXA",
                "SERVICO_FAIXA1_CLIENTES",
                "SERVICO_FAIXA2_TAXA",
                "SERVICO_FAIXA2_CLIENTES",
                "SERVICO_FAIXA3_TAXA",
                "SERVICO_FAIXA3_CLIENTES",
                "SERVICO_FAIXA4_TAXA",
                "SERVICO_FAIXA4_CLIENTES",
                "SERVICO_FATO_GERADOR"
            ]) + '\n')

            try:
                participantes = [x for x in ob.participantes]
            except:
                msg = "Não foi possível baixar as informações do diretório de participantes do Open Banking Brasil."
                if self.ignoraErros:
                    print(f'Erro: {msg}')
                    return

                raise BatchException(msg)

            totalParticipantes = len(participantes)
            totalProdutos = 0
            totalServicos = 0

            seqParticipante = 0
            seqProtudo = 0
            seqServico = 0

            for participante in participantes:
                seqProtudo = 0

                seqParticipante += 1
                try:
                    for produto in participante.produtos:
                        seqServico = 0

                        seqProtudo += 1
                        totalProdutos += 1
                        for servico in produto.servicos:
                            seqServico += 1
                            totalServicos += 1

                            row = [
                                str(hoje),
                                _fix(produto.endPoint),
                                _fix(seqParticipante),
                                _fix(participante.cnpj),
                                _fix(participante.nome),
                                _fix(seqProtudo),
                                _fix(produto.tipo),
                                _fix(produto.categoria),
                                _fix(produto.nome),
                                _fix(seqServico),
                                _fix(servico.nome),
                                _fix(servico.codigo),
                                _fix(servico.taxaMinima),
                                _fix(servico.taxaMaxima),
                                _fix(servico.faixa1Taxa),
                                _fix(servico.faixa1Clientes),
                                _fix(servico.faixa2Taxa),
                                _fix(servico.faixa2Clientes),
                                _fix(servico.faixa3Taxa),
                                _fix(servico.faixa3Clientes),
                                _fix(servico.faixa4Taxa),
                                _fix(servico.faixa4Clientes),
                                _fix(servico.fatoGerador),
                            ]
                            file.write(self.csvDelimiter.join(row) + '\n')
                except:
                    if self.ignoraErros:
                        print('Participante ignorado devido a erro no processamento de serviços de um produto.')
                        continue

                    raise BatchException("Um erro ocorreu durante o processamento de um participante.")

            file.flush()

        print('Processamento dos produtos concluído.')
        print(f"Total de participantes: {totalParticipantes}")
        print(f"Total de serviços: {totalServicos}")
        print('.')

    def pacotes_csv(self):
        """
        Realiza o download de todos os produtos disponíveis em todos participantes 
        no diretório do Open Banking Brasil, e salva os dados dos pacotes de serviços 
        em um arquivo no formato: `yyyymmdd_openbanking_pacotes.csv`.
        """

        ob = OpenBankingBR(
            cacheDir = self.cacheDir,
        )

        participantes: List[Participante]
        participantes = None

        hoje = date.today().strftime('%Y%m%d')

        fileName = os.path.join(self.dataDir, f'{hoje}_openbanking_pacotes.csv')
        if os.path.exists(fileName):
            os.remove(fileName)
            print(f'Arquivo {fileName} foi apagado para o reprocessamento.')

       
        with open(fileName, 'w', encoding=self.encoding) as file:
            file.write(self.csvDelimiter.join([
                "DATA_BASE",
                "API",
                "PARTICIPANTE_SEQ",
                "PARTICIPANTE_CNPJ",
                "PARTICIPANTE_NOME",
                "PRODUTO_SEQ",
                "PRODUTO_TIPO",
                "PRODUTO_CATEGORIA",
                "PRODUTO_NOME",
                "PACOTE_SEQ",
                "PACOTE_NOME",
                "PACOTE_TAXA_MINIMA",
                "PACOTE_TAXA_MAXIMA",
                "PACOTE_FAIXA1_TAXA",
                "PACOTE_FAIXA1_CLIENTES",
                "PACOTE_FAIXA2_TAXA",
                "PACOTE_FAIXA2_CLIENTES",
                "PACOTE_FAIXA3_TAXA",
                "PACOTE_FAIXA3_CLIENTES",
                "PACOTE_FAIXA4_TAXA",
                "PACOTE_FAIXA4_CLIENTES"
            ]) + '\n')

            try:
                participantes = [x for x in ob.participantes]
            except:
                msg = "Não foi possível baixar as informações do diretório de participantes do Open Banking Brasil."
                if self.ignoraErros:
                    print(f'Erro: {msg}')
                    return

                raise BatchException(msg)
        
            totalParticipantes = len(participantes)
            totalProdutos = 0
            totalPacotes = 0

            seqParticipante = 0
            seqProduto = 0
            seqPacote = 0

            for participante in participantes:
                seqParticipante += 1
                seqProduto = 0
                try:
                    for produto in participante.produtos:
                        seqPacote = 0
                        
                        totalProdutos += 1
                        seqProduto += 1
                        
                        for pacote in produto.pacotes:
                            totalPacotes += 1
                            seqPacote += 1
                            row = [
                                str(hoje),
                                _fix(produto.endPoint),
                                _fix(seqParticipante),
                                _fix(participante.cnpj),
                                _fix(participante.nome),
                                _fix(seqProduto),
                                _fix(produto.tipo),
                                _fix(produto.categoria),
                                _fix(produto.nome),
                                _fix(seqPacote),
                                _fix(pacote.nome),
                                _fix(pacote.taxaMinima),
                                _fix(pacote.taxaMaxima),
                                _fix(pacote.faixa1Taxa),
                                _fix(pacote.faixa1Clientes),
                                _fix(pacote.faixa2Taxa),
                                _fix(pacote.faixa2Clientes),
                                _fix(pacote.faixa3Taxa),
                                _fix(pacote.faixa3Clientes),
                                _fix(pacote.faixa4Taxa),
                                _fix(pacote.faixa4Clientes),
                            ]
                            file.write(self.csvDelimiter.join(row) + '\n')
                except:
                    if self.ignoraErros:
                        print('Participante ignorado devido a erro no processamento dos pacotes de serviço de um produto.')
                        continue

                    raise BatchException("Um erro ocorreu durante o processamento dos pacotes de serviço de um participante.")

            file.flush()

        print('Processamento dos pacotes de produtos concluído.')
        print(f"Total de participantes: {totalParticipantes}")
        print(f"Total de produtos: {totalProdutos}")
        print(f"Total de pacotes: {totalPacotes}")
        print('.')

def _fix(value) -> str:
    if value == None:
        return ""
    
    if type(value) == str:
        if value == 'NA':
            return ""

        # RFC-4180, paragraph "If double-quotes are used to enclose fields, then a double-quote 
        # appearing inside a field must be escaped by preceding it with another double quote."
        return "\"" + value.replace('\n', ' ').replace('\"', '\"\"') + "\""

    return str(value)

# %%



# -*- coding: utf-8 -*-

"""
OpenBankingBR 
~~~~~~~~~~~~~

Biblioteca para o consumo dos dados publicos do Open Banking Brasil.

Exemplo de uso:
```python
import openbankingbr as obr

```

:copyright: (c) 2021 by Gustavo J Knuppe.

:license: MIT.
"""

from .errors import (
    OpenBankingException,
    InvalidDataException,
)

from .__version__ import (
    __author__,
    __copyright__,
    __description__,
    __title__,
    __license__,
    __url__,
    __version__,
)

from .models import Participante
from .utils import fetchUrl

from typing import List

class OpenBankingBR():
    """
    Classe primária que possibilita listar todos os participantes cadastrados
    no diretório do OpenBanking Brasil.

    Exemplo:
    ```python
    ob = OpenBankingBR(cacheDir = './cache')
    for participante in ob.participantes:
        print(participante.nome)

    ```    
    """
    def __init__(self, cacheDir: str = None):
        self.cacheDir = cacheDir

    @property
    def participantes(self):
        """
        Gera uma lista com os participantes cadastrados no diretório do Open Banking Brasil.

        https://openbanking-brasil.github.io/areadesenvolvedor/#participantes-open-banking-brasil
        """
        
        data = fetchUrl('https://data.directory.openbankingbrasil.org.br/participants', self.cacheDir)
        if data == None:
            raise OpenBankingException("Não possível obter a lista de participantes do diretório oficial") from None
        elif not type(data) == list:
            raise InvalidDataException("Os dados retornados da lista de participantes não são válidos") from None

        for p in data:
            yield Participante(self, p)
    
    def busca_participante(
        self,
        cnpj: int = None,
        dominio: str = None,
        identificador: str = None,
        ):
        """
        Busca um participante utilizando apenas UM critério de pesquisa definido nos argumentos.

        > `cnpj` é o campo numérico tratado, originado do `RegistrationNumber`.

        > `dominio` é o dominio web, ex: `bcb.gov.br`

        > `identificador` é o identificador do participante `OrganisationId`.

        """
        for participante in self.participantes:
            if cnpj != None and participante.cnpj == cnpj:
                return participante
            elif identificador != None and participante.identificador:
                return participante
            elif dominio != None:
                for ep in participante.apiEndPoints:
                    if dominio in ep:
                        return participante
        
        return None

__all__ = [
    "batch",
    "errors",
    "models",
    "OpenBankingBR",
]



__title__ = 'openbankingbr'
__description__ = 'Python 3 package to extract public data for Open Banking Brasil.'
__url__ = 'https://github.com/knuppe/openbankingbr'
__version__ = '0.1.4'
__author__ = 'Gustavo J Knuppe'
__author_email__ = 'gustavoknuppe@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2021 Gustavo J Knuppe'



import os
import re
import platform
import time 
import requests
import json 

from requests.exceptions import (
    InvalidURL,
    Timeout
)

# Desabilita o warning
# InsecureRequestWarning: Unverified HTTPS request is being made to host
requests.packages.urllib3.disable_warnings()

from hashlib import md5
from datetime import date, datetime

from . import __version__
from .errors import (
    OpenBankingException,
    RequiredFieldException
)

def get(d: dict, key: str, required: bool = False, valueType: type = None):
    """
    Obtém dados de um objeto JSON usando a chave especificada.

    Caso especificado `valueType` será realizado o cast garantindo que o valor está
    no formato esperado. Caso o cast de algum erro e o `required` for `True`, será 
    retornado um erro.
    """
    cur = None
    try:
        keys = key.split('.')
        cur = d
        for k in keys:
            cur = cur[k]
            if cur is None:
                return None

        if valueType == None:
            return cur
        elif valueType == list:
            if type(cur) == list:
                return cur
            
            raise ValueError()
        elif valueType == dict:
            if type(cur) == dict:
                return cur

            raise ValueError()
        elif valueType == int:
            return int(cur)
        elif valueType == str:
            return str(cur)
        elif valueType == float:
            return float(cur)
        elif valueType == bool:
            return bool(cur)
        elif valueType == date:
            return datetime.fromisoformat(cur)
        elif valueType == datetime:
            # Haha! Na documentação falam que a data tem que ser seguir a especificação RFC-3339
            # porém, o próprio exemplo que dão não é uma data válida no formato...
            #
            # https://openbanking-brasil.github.io/areadesenvolvedor-fase1/#introducao-tipos-de-dados-comuns
            #
            if type(cur) == str:
                cur = str(cur).rstrip('Z')
                
            return datetime.fromisoformat(cur)
        else:
            raise NotImplementedError("Tipo de dados solicitado é desconhecido")
        
    except ValueError:
        if required:
            raise RequiredFieldException(f"O tipo do valor atual '{type(cur)}' não está no formato esperado '{valueType}'") from None
    except KeyError:
        pass
        if required:
            raise RequiredFieldException(f'Campo obrigatório não encontrado nos dados: {key}') from None
    except Exception as e:
        if required:
            raise e from None

    return None

def getInt(dados: dict, key: str) -> int:
    """
    Procura a chave especificada nos dados, caso encontrado algum valor é extraido somente
    os caracteres reconhecidos como digitos então é retornado um cast de int do valor.

    Caso ocorra um erro ou o valor seja inválido, será retornado `None`.

    Essa função é usada pois em vários campos as financeiras formatam os valores, mesmo
    com a documentação orientando a não formatar. 
    """
    value = get(dados, key, valueType=str, required=False)
    if value == None:
        return None

    try:
        value = re.sub(r'\D', '', value)
        return int(value)
    except:
        return None

def fetchUrl(url: str, cacheDir: str = None) -> dict:
    """
    Executa a requisição dos dados da URL.

    Caso especificado o cacheDir, os dados de cada request serão armazenado sem arquivos JSON
    de cache por um dia.
    """
    if cacheDir != None:
        if not os.path.exists(cacheDir):
            os.makedirs(cacheDir)
        elif not os.path.isdir(cacheDir):
            raise OpenBankingException("O cacheDir especificado não é um diretório.")

        cacheEntry = os.path.join(cacheDir, md5(url.encode('ascii')).hexdigest() + '.json')
        if file_was_created_today(cacheEntry):
            with open(cacheEntry, 'r', encoding='utf8') as f:
                return json.load(f)
        else:
            if os.path.exists(cacheEntry):
                os.remove(cacheEntry)
            
    try:
        res = requests.get(
            url, 
            headers = {
                'User-Agent': f'OpenBankingBR/{__version__}; (+https://github.com/knuppe/openbankingbr)'
            },
            verify=False  # Lógico que como estamos no Brasil, financeiras estão cagando para ter um certificado válido nos endpoints... 
        )

        if not res.ok:
           return None

        data = res.json()

        if data != None:
            with open(cacheEntry, 'w', encoding='utf8') as f:
                json.dump(data, f, ensure_ascii=False)

        return data
    except InvalidURL:
        # ignora pois o erro é meio grotesco.
        return None
    except ValueError:
        raise OpenBankingException("O conteúdo retornado pelo endpoint não possui um JSON válido.") from None
    except:
        return None

def fetchUrlPages(url: str, cacheDir: str = None) -> dict:
    """
    Realiza a requisição dos dados da API retornando os dados paginados.

    https://openbanking-brasil.github.io/areadesenvolvedor-fase1/#introducao-paginacao
    """
    root = fetchUrl(url, cacheDir = cacheDir)
    if root == None or not type(root) == dict:
        return None
    
    if not 'data' in root:
        return None

    data = root['data']
    data['endPoint'] = url

    yield data

    # Knuppe: Olha, poderia usar o total de paginas que tá no meta... mas
    #         como sei que as financeiras vão fazer api bugada, vou olhar 
    #         somente se existe o link de próxima pagina e boa.
    #
    # totalPages = get(root, 'meta.totalPages', valueType=int, required=False)
    # totalPages = 1 if totalPages is None else totalPages
    # for page in range(2, totalPages):

    nextPage = get(root, 'links.next', valueType=str, required = False)
    if nextPage is None:
        return
    
    if nextPage.startswith('http://') or nextPage.startswith('https://'):
        yield fetchUrlPages(nextPage, cacheDir = cacheDir)
    

def file_was_created_today(file: str) -> bool:
    if not os.path.exists(file):
        return False

    ctime: datetime
    
    if platform.system() == "Windows":
        ctime = datetime.fromtimestamp(os.path.getctime(file))
    else:
        stat = os.stat(file)
        try:
            ctime = datetime.fromtimestamp(stat.st_birthtime)
        except:
            ctime = datetime.fromtimestamp(stat.st_mtime)

    return ctime.date() == date.today()



from .errors import OpenBankingException, InvalidDataException
from .utils import get, getInt, fetchUrlPages
from .spec import nomeProdutos, chaveCategorias

from typing import List, Iterator, Generator
from datetime import date

import re 

class DataObject():
    """
    Objeto base com informações dos dados retornados de uma api do Open Banking Brasil.
    """
    def __init__(self, openBaking, dados: dict):
        if not isinstance(dados, dict):
            raise InvalidDataException("O parâmetro de dados do objeto não é um dicionário")

        self.openBaking = openBaking
        self.dados      = dados

    @property
    def endPoint(self) -> str:
        """
        Retorna o end-point da API onde os dados do Open Banking foram obtidos.
        """
        if 'endPoint' in self.dados:
            return self.dados['endPoint']
        
        return None

class Participante(DataObject):
    """
    Representa uma empresa participante listada no diretório do Open Banking Brasil.
    """

    def __init__(self, openBaking, data: dict):
        DataObject.__init__(self, openBaking, data)
        
        if not 'AuthorisationServers' in data:
            raise InvalidDataException("AuthorisationServers não está definido nos dados do participante.")
        elif type(data['AuthorisationServers']) != list:
            raise InvalidDataException("AuthorisationServers não é uma lista")

        self._endpoints = list()
        for server in data['AuthorisationServers']:
            if 'ApiResources' in server and isinstance(server['ApiResources'], list):
                for api in server['ApiResources']:
                    if 'ApiDiscoveryEndpoints' in api and isinstance(api['ApiDiscoveryEndpoints'], list):
                        for endPoint in api['ApiDiscoveryEndpoints']: 
                            if 'ApiEndpoint' in endPoint and type(endPoint['ApiEndpoint']) == str:
                                self._endpoints.append(endPoint['ApiEndpoint'])

    @property
    def nome(self) -> str:
        """
        Retorna o nome (OrganisationName) do participante.
        """
        return str(self.dados['OrganisationName'])

    @property
    def cnpj(self) -> int:
        """
        Retorna o CNPJ numérico com base no `RegistrationNumber` do participante.
        """
        res = re.sub(r'[^\d]+', '', self.dados['RegistrationNumber'])
        return 0 if not res else int(res)        

    @property
    def registrationId(self) -> str:
        """
        Retorna a identificação de registro do participante.
        """
        return str(self.dados['RegistrationId'])
    
    @property
    def registrationNumber(self) -> str:
        """
        Retorna o "número" de registro no participante (CNPJ) no formato que a API retorna.

        Note que existem algumas instituições que mandam o CNPJ formatado. Por este motivo existe 
        a propriedade `.cnpj` que retorna um `int` com o valor tratado.
        """
        return get(self.dados, 'RegistrationNumber', valueType=str, required=True)

    @property
    def identificador(self) -> str:
        """
        Retorna o identificador único do participante.
        """
        return get(self.dados, 'OrganisationId', valueType=str, required=True)

    @property
    def apiEndPoints(self) -> List[str]:
        """
        Retorna uma lista com todos endereços endpoint das API expostas pelo participante.
        """
        return self._endpoints

    @property
    def agencias(self):
        """
        Gera uma lista com todas as agencias (postos de atendimento) do participante no Open Banking Brasil.

        Exemplo de uso:
        ```python
        for agencia in participante.agencias:
            print(agencia.nome)
        ```        
        """
        for endPoint in self._endpoints:
            if not endPoint.endswith('/open-banking/channels/v1/branches'):
                continue

            for page in fetchUrlPages(endPoint, self.openBaking.cacheDir):
                if not 'brand' in page:
                    continue
            
                brand = page['brand']
                if not 'companies' in brand:
                    continue

                for company in brand['companies']:
                    branches = get(company, 'branches', valueType=list, required=True)

                    for branch in branches:
                        branch['endPoint'] = page['endPoint']

                        yield Agencia(self, branch)

            break                         

    @property
    def produtos(self):
        """
        Gera uma lista com todos os produtos oferecidos pelo participante.

        Exemplo de uso:
        ```python
        for produto in participante.produtos:
            print(produto)
        ```
        """

        # 'personal-accounts', 'business-accounts', 'personal-credit-cards', 'business-credit-cards', 'personal-loans', 'business-loans', 'personal-financings', 'business-financings', 'personal-invoice-financings', 'business-invoice-financings', 'personal-unarranged-account-overdraft', 'business-unarranged-account-overdraft', 

        for endPoint in self._endpoints:
            if not '/open-banking/products-services/v1/' in endPoint:
                continue

            keys = [
                "personalAccounts",
                "businessAccounts",
                "personalCreditCards",
                "businessCreditCards",
                "personalLoans",
                "businessLoans",
                "personalFinancings",
                "businessFinancings",
                "personalInvoiceFinancings",
                "businessInvoiceFinancings",
                "personalUnarrangedAccountOverdraft",
                "businessUnarrangedAccountOverdraft",
            ]

            for page in fetchUrlPages(endPoint, self.openBaking.cacheDir):
                if not 'brand' in page:
                    continue
            
                brand = page['brand']
                if not 'companies' in brand:
                    continue

                key = None
                for company in brand['companies']:
                    for k in keys:
                        if not k in company:
                            continue
                        
                        key = k
                        break

                    if key is None:
                        continue

                    if not type(company[key]) == list:
                        continue

                    for item in company[key]:
                        item['endPoint'] = page['endPoint']
                        
                        if 'interestRates' in item and type(item['interestRates']) == list:
                            interestSequence = 0
                            for interestRate in item['interestRates']:
                                interestSequence += 1
                                interestRate['interestSequence'] = interestSequence
                                yield Produto(self, key, item, interestRate)

                        elif 'interest' in item and type(item['interest']) == dict:             # falta de padrão da api é foda...
                            if 'rates' in item['interest'] and type(item['interest']['rates']) == list:
                                interestSequence = 0
                                for interestRate in item['interest']['rates']:
                                    interestSequence += 1
                                    interestRate['interestSequence'] = interestSequence
                                    yield Produto(self, key, item, interestRate)
                        else:
                            yield Produto(self, key, item)
    @property
    def status(self) -> str:
        """
        Retorna o status do registro do diretório da organização.
        """
        return get(self.dados, 'Status', valueType=str, required=True)

class Agencia(DataObject):
    """
    Represnta uma agencia (Posto de Atendimento) de um participante do Open Banking Brasil.
    """

    def __init__(self, participante, data: dict):
        DataObject.__init__(self, participante.openBaking, data)

        assert(isinstance(participante, Participante))

        self._participante = participante

    @property
    def tipo(self) -> str:
        """
        Retorna o tipo da agência, segundo a regulamentação do Bacen, na Resolução Nº 4072, de 26 de
        abril de 2012: agência de instituições financeiras e demais instituições, autorizadas
        a funcionar pelo Banco Central do Brasil, destinada à prática das atividades para as 
        quais a instituição esteja regularmente habilitada.
        """
        return get(self.dados, 'identification.type', valueType=str)

    @property
    def codigo(self) -> int:
        """
        Retorna o código identificador da agência. Ex. '3006','3035', '1382', '2516', '2856'.
        """
        return getInt(self.dados, 'identification.code')

    @property
    def digitoVerificador(self) -> str:
        """
        Retorna o dígito verificador do código da agência.
        """
        digito = get(self.dados, 'identification.checkDigit', valueType=str)
        return None if digito == 'NA' else digito

    @property
    def nome(self) -> str:
        """
        Retorna o nome da agência.
        """
        return get(self.dados, 'identification.name', valueType=str)

    @property
    def servicos(self) -> List[str]:
        """
        Retorna uma lista com os códigos dos serviços oferecidos na agência.
        """
        items = list()
        servicecs = get(self.dados, 'services', valueType=list, required=False)
        if servicecs != None:
            for service in servicecs:
                code = get(service, 'code', valueType=str)
                if code:
                    items.append(code)

        return items
    
    @property
    def acessoAoPublico(self) -> bool:
        """
        Indica se a instalação da agência tem acesso restrito a clientes. Se `False` o acesso é restrito aos clientes.
        """
        return get(self.dados, 'availability.isPublicAccessAllowed', valueType=bool) == True

    @property
    def endereco(self) -> str:
        """
        Retorna o endereço onde a agência está localizada.
        """
        return get(self.dados, 'postalAddress.address', valueType=str, required=False)

    @property
    def enderecoComplemento(self) -> str:
        """
        Retorna o complemento do endereço onde a agência está localizada.
        """
        valor = get(self.dados, 'postalAddress.additionalInfo', valueType=str, required=False)
        return None if valor == 'NA' else valor

    @property
    def cep(self) -> int:
        """
        Retorna o CEP onde a agência está localizada.
        """
        return getInt(self.dados, 'postalAddress.postCode')

    @property
    def cidade(self) -> str:
        """
        Retorna a cidade onde a agência está localizada.
        """
        return get(self.dados, 'postalAddress.townName', valueType=str, required=False)

    @property
    def uf(self) -> str:
        """
        Retorna o estado onde a agência está localizada.
        """
        return get(self.dados, 'postalAddress.countrySubDivision', valueType=str, required=False)

    @property
    def bairro(self) -> str:
        """
        Retorna o bairro onde a agência está localizada.
        """
        return get(self.dados, 'postalAddress.districtName', valueType=str, required=False)

    @property
    def codigoIBGE(self) -> int:
        """
        Retorna o código do IBGE da cidade onde a agência está localizada.
        """
        return getInt(self.dados, 'postalAddress.ibgeCode')

    @property
    def latitude(self) -> float:
        """
        Retorna a latitude onde a agência está localizada.
        """
        return get(self.dados, 'postalAddress.geographicCoordinates.latitude', valueType=float, required=False)

    @property
    def longitude(self) -> float:
        """
        Retorna a longitude onde a agência está localizada.
        """
        return get(self.dados, 'postalAddress.geographicCoordinates.longitude', valueType=float, required=False)

    @property
    def telefone(self) -> str:
        """
        Retorna o telefone da agência.

        Note que este campo é construído com regras priorizando um telefone fixo depois telefone móvel.
        """
        phones = get(self.dados, 'phones', valueType=list, required=False)
        if phones is None:
            return None
        
        for tipo in ['FIXO', 'MOVEL']:
            for phone in phones:
                if get(phone, 'type') == tipo:
                    valor = ""
                    if 'countryCallingCode' in phone:
                        i = getInt(phone, 'countryCallingCode')
                        if i != None:
                            valor += i.__str__()

                    if 'areaCode' in phone:
                        i = getInt(phone, 'areaCode').__str__()
                        if i != None:
                            valor += i.__str__()
                    
                    if 'number' in phone:
                        i = getInt(phone, 'number').__str__()
                        if i != None:
                            valor += i.__str__()
                
                    if len(valor) > 3:
                        return valor

        return None


class Servico(DataObject):
    """
    Representa o detalhamento de uma tarifa de serviço cobrada em um produto.
    """
    def __init__(self, openBaking, dados: dict):
        DataObject.__init__(self, openBaking, dados)

        self._taxa     = [None, None, None, None]
        self._clientes = [None, None, None, None]

        if get(dados, 'prices', valueType=list, required=False) != None:
            for price in dados['prices']:
                if price['interval']    == '1_FAIXA':
                    self._taxa[0]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[0] = get(price, 'customers.rate', valueType=float, required=False)
                elif price['interval']  == '2_FAIXA':
                    self._taxa[1]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[1] = get(price, 'customers.rate', valueType=float, required=False)
                elif price['interval']  == '3_FAIXA':
                    self._taxa[2]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[2] = get(price, 'customers.rate', valueType=float, required=False)
                elif price['interval']  == '4_FAIXA':
                    self._taxa[3]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[3] = get(price, 'customers.rate', valueType=float, required=False)

    @property
    def nome(self) -> str:
        """
        Obtém o nome do serviço.
        """
        return get(self.dados, 'name', valueType=str, required=True)

    @property
    def codigo(self) -> str:
        """
        Obtém o código do serviço.
        """
        return get(self.dados, 'code', valueType=str, required=False)

    @property
    def fatoGerador(self) -> str:
        """
        Fatos geradores de cobrança que incidem sobre o produto.
        """
        return get(self.dados, 'chargingTriggerInfo', valueType=str, required=False)

    @property
    def faixa1Taxa(self) -> float:
        """
        Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste serviço.

        Segundo Normativa nº 32, BCB, de 2020: Distribuição de frequência relativa dos valores de tarifas 
        cobradas dos clientes, de que trata o § 2º do art. 3º da Circular nº 4.015, de 2020, deve dar-se 
        com base em quatro faixas de igual tamanho, com explicitação dos valores sobre a mediana em cada 
        uma dessas faixas. Informando: 1ª faixa, 2ª faixa , 3ª faixa e 4ª faixa
        """
        return self._taxa[0]

    @property
    def faixa2Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste serviço."""
        return self._taxa[1]

    @property
    def faixa3Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste serviço."""
        return self._taxa[2]

    @property
    def faixa4Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste serviço."""
        return self._taxa[3]

    @property
    def faixa1Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 1."""
        return self._clientes[0]

    @property
    def faixa2Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 2."""
        return self._clientes[1]

    @property
    def faixa3Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 3."""
        return self._clientes[2]

    @property
    def faixa4Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 4."""
        return self._clientes[3]

    @property
    def taxaMaxima(self) -> float:
        """Taxa máxima cobrada pelo serviço."""
        return get(self.dados, 'maximum.value', valueType=float)

    @property
    def taxaMinima(self) -> float:
        """Taxa mínima cobrada pelo serviço."""
        return get(self.dados, 'minimum.value', valueType=float)



class Pacote(DataObject):
    """
    Representa um pacote de serviços oferecido para um produto.
    """
    def __init__(self, openBaking, dados: dict):
        DataObject.__init__(self, openBaking, dados)

        self._taxa     = [None, None, None, None]
        self._clientes = [None, None, None, None]

        if get(dados, 'prices', valueType=list, required=False) != None:
            for price in dados['prices']:
                if price['interval']    == '1_FAIXA':
                    self._taxa[0]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[0] = get(price, 'customers.rate', valueType=float, required=False)
                elif price['interval']  == '2_FAIXA':
                    self._taxa[1]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[1] = get(price, 'customers.rate', valueType=float, required=False)
                elif price['interval']  == '3_FAIXA':
                    self._taxa[2]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[2] = get(price, 'customers.rate', valueType=float, required=False)
                elif price['interval']  == '4_FAIXA':
                    self._taxa[3]     = get(price, 'value'         , valueType=float, required=False)
                    self._clientes[3] = get(price, 'customers.rate', valueType=float, required=False)

    @property
    def nome(self) -> str:
        """
        Obtem o nome do pacote de serviços.
        """
        return get(self.dados, 'name', valueType=str, required=True)

    @property
    def faixa1Taxa(self) -> float:
        """
        Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste pacote.

        Segundo Normativa nº 32, BCB, de 2020: Distribuição de frequência relativa dos valores de tarifas 
        cobradas dos clientes, de que trata o § 2º do art. 3º da Circular nº 4.015, de 2020, deve dar-se 
        com base em quatro faixas de igual tamanho, com explicitação dos valores sobre a mediana em cada 
        uma dessas faixas. Informando: 1ª faixa, 2ª faixa , 3ª faixa e 4ª faixa
        """
        return self._taxa[0]

    @property
    def faixa2Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste pacote."""
        return self._taxa[1]

    @property
    def faixa3Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste pacote."""
        return self._taxa[2]

    @property
    def faixa4Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste pacote."""
        return self._taxa[3]

    @property
    def faixa1Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 1."""
        return self._clientes[0]

    @property
    def faixa2Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 2."""
        return self._clientes[1]

    @property
    def faixa3Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 3."""
        return self._clientes[2]

    @property
    def faixa4Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 4."""
        return self._clientes[3]

    @property
    def taxaMaxima(self) -> float:
        """Taxa máxima cobrada pelo pacote."""
        return get(self.dados, 'maximum.value', valueType=float)

    @property
    def taxaMinima(self) -> float:
        """Taxa mínima cobrada pelo pacote."""
        return get(self.dados, 'minimum.value', valueType=float)


class Produto(DataObject):
    """
    Representa um produto negociado por um participante do Open Banking Brasil.
    """
    def __init__(self, participante, key: str, data: dict, interestRate: dict = None):
        DataObject.__init__(self, participante.openBaking, data)

        assert(isinstance(participante, Participante))

        self._key = key
        self._participante = participante
        self._taxa       = [None, None, None, None]
        self._jurosRate   = None
        self._clientes  = [None, None, None, None]
        self._jutosTipo   = None
        self._minimumRate = None 
        self._maximumRate = None

        if interestRate != None:
            self._jurosRate = get(interestRate, "rate", valueType=float, required=False)
            self._jutosTipo = get(interestRate, "referentialRateIndexer", valueType=str, required=False)

            if self._jutosTipo == 'NA':
                self._jutosTipo = None

            if 'applications' in interestRate:
                for app in interestRate['applications']:
                    if app['interval']    == '1_FAIXA':
                        self._taxa[0]     = get(app, 'indexer.rate'  , valueType=float, required=False)
                        self._clientes[0] = get(app, 'customers.rate', valueType=float, required=False)
                    elif app['interval']  == '2_FAIXA':
                        self._taxa[1]     = get(app, 'indexer.rate'  , valueType=float, required=False)
                        self._clientes[1] = get(app, 'customers.rate', valueType=float, required=False)
                    elif app['interval']  == '3_FAIXA':
                        self._taxa[2]     = get(app, 'indexer.rate'  , valueType=float, required=False)
                        self._clientes[2] = get(app, 'customers.rate', valueType=float, required=False)
                    elif app['interval']  == '4_FAIXA':
                        self._taxa[3]     = get(app, 'indexer.rate'  , valueType=float, required=False)
                        self._clientes[3] = get(app, 'customers.rate', valueType=float, required=False)

            self._minimumRate = get(interestRate, 'minimumRate', valueType=float, required=False)
            self._maximumRate = get(interestRate, 'maximumRate', valueType=float, required=False)

    @property
    def nome(self) -> str:
        """
        Retorna o nome associado ao produto.
        """
        if self._key in ['personalUnarrangedAccountOverdraft', 'businessUnarrangedAccountOverdraft']:
            return 'Adiantamento a Depositante'

        if 'type' in self.dados:
            productType = self.dados['type']

            if productType in nomeProdutos:
                return nomeProdutos[productType]
        
        if 'name' in self.dados and type(self.dados['name']) == str:
            return self.dados['name']

        return None

    @property
    def tipo(self) -> str:
        """
        Retorna o tipo do produto.
        """
        if self._key in ['personalCreditCards', 'businessCreditCards']:
            return get(self.dados, 'identification.product.type', required=True, valueType=str)

        if 'type' in self.dados and type(self.dados['type']) == str:
            return self.dados['type']

        if self._key in ['personalUnarrangedAccountOverdraft', 'businessUnarrangedAccountOverdraft']:
            return 'ADP'

        return 'UNKNOWN'

    @property
    def categoria(self) -> str:
        """
        Retorna uma categoria do produto calculada internamente na biblioteca.
        """
        if self._key in chaveCategorias:
            return chaveCategorias[self._key]
        
        return 'Outros'

    @property
    def indexador(self) -> str:
        """
        Retorna o tipo da taxa referencial ou indedxador.
        """
        return self._jutosTipo

    @property
    def indexadorRate(self) -> float:
        """
        Percentual que incide sobre a composição das taxas de juros remuneratórios.
        """
        return self._jurosRate

    @property
    def jurosSeq(self) -> int:
        """
        Retorna um campo sequencial que identifica qual o valor sequencial de juros 
        é os 
        """
        seq = get(self.dados, 'interestSequence', valueType=int, required=False)
        return 1 if seq is None else seq

    @property
    def faixa1Taxa(self) -> float:
        """
        Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste produto.

        Segundo Normativa nº 32, BCB, de 2020: Distribuição de frequência relativa dos valores de tarifas 
        cobradas dos clientes, de que trata o § 2º do art. 3º da Circular nº 4.015, de 2020, deve dar-se 
        com base em quatro faixas de igual tamanho, com explicitação dos valores sobre a mediana em cada 
        uma dessas faixas. Informando: 1ª faixa, 2ª faixa , 3ª faixa e 4ª faixa
        """
        return self._taxa[0]

    @property
    def faixa2Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste produto."""
        return self._taxa[1]

    @property
    def faixa3Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste produto."""
        return self._taxa[2]

    @property
    def faixa4Taxa(self) -> float:
        """Percentual que corresponde a mediana da taxa efetiva cobrada do cliente pela contratação deste produto."""
        return self._taxa[3]

    @property
    def faixa1Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 1."""
        return self._clientes[0]

    @property
    def faixa2Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 2."""
        return self._clientes[1]

    @property
    def faixa3Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 3."""
        return self._clientes[2]

    @property
    def faixa4Clientes(self) -> float:
        """Percentual de clientes com a taxa descrita na faixa 4."""
        return self._clientes[3]

    @property
    def participante(self) -> Participante:
        """Retorna o participante que oferece este produto."""
        return self._participante

    @property
    def taxaMaxima(self) -> float:
        """Taxa máxima efetiva dos contratos deste produto."""
        return self._maximumRate

    @property
    def taxaMinima(self) -> float:
        """Taxa mínima efetiva dos contratos deste produto.."""
        return self._minimumRate

    @property
    def rede(self) -> bool:
        """
        Retorna a rede ao qual o produto pertence.
        """
        return get(self.dados, "identification.creditCard.network", valueType=str, required=False)

    @property
    def programaRecompensas(self) -> bool:
        """Retorna `True` se o produto possui programa de recompensas."""
        return get(self.dados, "rewardsProgram.hasRewardProgram", valueType=bool, required=False)

    @property
    def servicos(self) -> Generator[Servico, None, None]:
        """
        Gera uma lista com o detalhamento das tarifas de serviço cobradas no produto.
        """
        if 'fees' in self.dados and type(self.dados['fees']) == dict:
            fees = self.dados['fees']
            for key in ['priorityServices', 'otherServices', 'services']:
                if key in fees and type(fees[key]) == list:
                    for service in fees[key]:
                        yield Servico(self.openBaking, service)

    @property
    def pacotes(self) -> Generator[Servico, None, None]:
        """
        Gera uma lista com os pacotes de serviços ofertados para o produto.
        """
        if 'serviceBundles' in self.dados and type(self.dados['serviceBundles']) == list:
            for bundle in self.dados['serviceBundles']:
                if type(bundle) == dict:
                    yield Pacote(self.openBaking, bundle)

    def __repr__(self):
        return f'Produto {self.tipo} > {self.nome}'






class OpenBankingException(Exception):
    """
    Error raised when problems arise
    """

class BatchException(OpenBankingException):
    """
    Ocorre devido a um erro no processamento batch de dados públicos.
    """

class InvalidDataException(OpenBankingException):
    """
    Ocorre quando algum dado inválido ou não esperado foi encontrado.
    """

class RequiredFieldException(OpenBankingException):
    """
    Ocorre quando um campo obrigatório não é encontrado nos dados do open-banking de um participante.
    """


import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

info = dict()
with open('src/openbankingbr/__version__.py', 'r') as f:
    lines = f.readlines()
    for line in lines:
        if ' = ' in line:
            key = line.split(' = ')[0]
            val = line.split(' = ')[1]
            val = val.rstrip('\t\r\n\'').lstrip('\t\'')
            info[key] = val

setuptools.setup(
    name="openbankingbr",
    version=info['__version__'],
    author=info['__author__'],
    author_email=info['__author_email__'],
    description=info['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/knuppe/openbankingbr",
    project_urls={
        "Bug Tracker": "https://github.com/knuppe/openbankingbr/issues",
    },    
    license='MIT',
    packages=setuptools.find_packages(where="src"),    
    install_requires=[
        "requests"
    ],
    keywords=['openbanking', 'bacen', 'banco-central-brasil'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Portuguese (Brazilian)",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    python_requires=">=3.6",
)




nomeProdutos = {
    'CONTA_DEPOSITO_A_VISTA'                                        : 'Conta corrente',
    'CONTA_POUPANCA'                                                : 'Conta poupança',
    'CONTA_PAGAMENTO_PRE_PAGA'                                      : 'Conta de pagamento pré-paga',
    'DESCONTO_DUPLICATAS'                                           : 'Desconto de duplicatas',
    'DESCONTO_CHEQUES'                                              : 'Desconto de cheques',
    'ANTECIPACAO_FATURA_CARTAO_CREDITO'                             : 'Antecipação de fatura de cartão de crédito',
    'OUTROS_DIREITOS_CREDITORIOS_DESCONTADOS'                       : 'Outros direitos creditórios descontados',
    'OUTROS_TITULOS_DESCONTADOS'                                    : 'Outros títulos descontados',
    'EMPRESTIMO_CREDITO_PESSOAL_CONSIGNADO'                         : 'Crédito pessoal consignado',
    'EMPRESTIMO_CREDITO_PESSOAL_SEM_CONSIGNACAO'                    : 'Crédito pessoal sem consignação',
    'EMPRESTIMO_HOME_EQUITY'                                        : 'Home equity',
    'EMPRESTIMO_MICROCREDITO_PRODUTIVO_ORIENTADO'                   : 'Microcrédito produtivo orientado',
    'EMPRESTIMO_CHEQUE_ESPECIAL'                                    : 'Cheque especial',
    'EMPRESTIMO_CONTA_GARANTIDA'                                    : 'Conta garantida',
    'EMPRESTIMO_CAPITAL_GIRO_PRAZO_VENCIMENTO_ATE_365_DIAS'         : 'Capital de giro com prazo de vencimento até 365 dias',
    'EMPRESTIMO_CAPITAL_GIRO_PRAZO_VENCIMENTO_SUPERIOR_365_DIAS'    : 'Capital de giro com prazo de vencimento superior a 365 dias',
    'EMPRESTIMO_CAPITAL_GIRO_ROTATIVO'                              : 'Capital de giro rotativo',
    'FINANCIAMENTO_AQUISICAO_BENS_VEICULOS_AUTOMOTORES'             : 'Aquisição de bens - Veículos automotores',
    'FINANCIAMENTO_AQUISICAO_BENS_OUTROS_BENS'                      : 'Aquisição de bens - Outros bens',
    'FINANCIAMENTO_MICROCREDITO'                                    : 'Financimento microcrédito',
    'FINANCIAMENTO_RURAL_CUSTEIO'                                   : 'Financimento rural - Custeio',
    'FINANCIAMENTO_RURAL_INVESTIMENTO'                              : 'Financimento rural - Investimento',
    'FINANCIAMENTO_RURAL_COMERCIALIZACAO'                           : 'Financimento rural - Comercialização',
    'FINANCIAMENTO_RURAL_INDUSTRIALIZACAO'                          : 'Financimento rural - Industrialização',
    'FINANCIAMENTO_IMOBILIARIO_SISTEMA_FINANCEIRO_HABITACAO_SFH'    : 'Financimento imobiliário - Sistema Financeiro da Habitação (SFH)',
    'FINANCIAMENTO_IMOBILIARIO_SISTEMA_FINANCEIRO_HABITACAO_SFI'    : 'Financimento imobiliário - Sistema Financeiro da Imobiliário (SFI)',

    # Fora do padrão do bacen -
    
    # BANCO ORIGINAL - Tem vários outros fora do padrão. Não vou tratar o lixo todo...
    'CONTA_PAGAMENTO'                                               : 'Conta salário',
    'EMPRESTIMO_CREDITO_PESSOAL'                                    : 'Crédito pessoal sem consignação',
    'EMPRESTIMO_CREDITO_PESSOAL_NAO_CONSIGNADO'                     : 'Crédito pessoal sem consignação',
    
    # Bradesco
    'FINANCIAMENTO_IMOBILIARIO_SISTEMA_FINANCEIRO_HABITACAO-SFH'    : 'Financimento imobiliário - Sistema Financeiro da Habitação (SFH)',
    'FINANCIAMENTO_IMOBILIARIO_SISTEMA_FINANCEIRO_HABITACAO-SFI'    : 'Financimento imobiliário - Sistema Financeiro da Imobiliário (SFI)',
}

chaveCategorias = {
    "personalAccounts"                                              : "Conta Corrente",
    "businessAccounts"                                              : "Conta Corrente",
    "personalCreditCards"                                           : "Cartão de Crédito",
    "businessCreditCards"                                           : "Cartão de Crédito",
    "personalLoans"                                                 : "Empréstimo",
    "businessLoans"                                                 : "Empréstimo",
    "personalFinancings"                                            : "Financiamento",
    "businessFinancings"                                            : "Financiamento",
    "personalInvoiceFinancings"                                     : "Antecipação de Recebíveis",
    "businessInvoiceFinancings"                                     : "Antecipação de Recebíveis",
    "personalUnarrangedAccountOverdraft"                            : "Adiantamento a Depositante",
    "businessUnarrangedAccountOverdraft"                            : "Adiantamento a Depositante",
}